/*
 * Plot - plot time-based data on screen and to file with interactive controls
 * Copyright (C) 2006  Jonas Berlin <xkr47@outerspace.dyndns.org>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ****
 * data_scrollbar_controller - functions for managing a set of gtk
 * scrollbars for selecting a subrange from a given maximum range (for
 * example x axis of a plot).
 */

#include <glade/glade.h>
#include <gtk/gtk.h>
#include <math.h>
#include <time.h>
#include <stdio.h>

#include "data_scrollbar_controller.xh"
#include "mem.xh"
#include "util.xh"

#define THIS data_scrollbar_controller_t *this = (data_scrollbar_controller_t*)user_data;

static void scroll_off_rel__value_changed(GtkRange *range, gpointer user_data) {
   THIS;
   gtk_range_set_value(range, 0.0);
   this->scroll_off_rel_pos = NAN;
}

static gboolean scroll_off_rel__change_value(GtkRange *range, GtkScrollType scroll, gdouble value, gpointer user_data) {
   THIS;

   if(__isnan(this->scroll_off_rel_pos)) {
      this->scroll_off_rel_pos = gtk_range_get_value(this->scroll_off);
   }
   if(scroll == GTK_SCROLL_JUMP) {
      GtkAdjustment *hsa_off = gtk_range_get_adjustment(this->scroll_off);
      gtk_range_set_value(this->scroll_off, this->scroll_off_rel_pos + value * hsa_off->page_size);
   } else {
      double off;
      switch((int)scroll) {
	 case GTK_SCROLL_PAGE_FORWARD:  off =  1;    break;
	 case GTK_SCROLL_PAGE_BACKWARD: off = -1;    break;
	 case GTK_SCROLL_STEP_FORWARD:  off =  0.25; break;
	 case GTK_SCROLL_STEP_BACKWARD: off = -0.25; break;
	 default:
	    return 0;
      }
      GtkAdjustment *hsa_off = gtk_range_get_adjustment(this->scroll_off);
      gtk_range_set_value(this->scroll_off, gtk_range_get_value(this->scroll_off) + off * hsa_off->page_size);
   }

   return 0; // continue
}

static void scroll_off__value_changed(GtkRange *range, gpointer user_data) {
   THIS;

   double o = gtk_range_get_value(this->scroll_off);
   double v = exp(gtk_range_get_value(this->scroll_scale));
   if(this->label_off_min) gtk_label_set_text(this->label_off_min, format_time((my_time_t)o, 1));
   if(this->label_off_max) gtk_label_set_text(this->label_off_max, format_time((my_time_t)(o + v), 1));

   this->num_changes++;
}

static void scroll_scale__value_changed(GtkRange *range, gpointer user_data) {
   THIS;

   double newval = exp(gtk_range_get_value(this->scroll_scale));

   GtkAdjustment *hsa_off = gtk_range_get_adjustment(this->scroll_off);

   // adjust off
   if(this->scale) {
      double oldval = this->old_scale_value;
      double old_off = hsa_off->value;
      double old_scale = gtk_range_get_value(this->scale);
      double new_off = old_off - (newval - oldval) * old_scale;
      gtk_adjustment_set_value(hsa_off, new_off);
      this->old_scale_value = newval;

      if(new_off < hsa_off->lower) {
	 double new_scale = old_scale + (new_off - hsa_off->lower) / newval;
	 gtk_range_set_value(this->scale, new_scale);
      }
      if(new_off + newval > hsa_off->upper) {
	 double new_scale = old_scale + (new_off + newval - hsa_off->upper) / newval;
	 gtk_range_set_value(this->scale, new_scale);
      }
   }

   // make sure off + scale don't exceed max
   hsa_off->page_size = newval;
   if(hsa_off->value + newval > hsa_off->upper) {
      gtk_adjustment_set_value(hsa_off, hsa_off->upper - newval);
   }
   // tell the guys we are finished changing
   gtk_adjustment_changed(hsa_off);

   // ##TODO## assumes scrollbar unit is time
   if(this->entry_scale && !GTK_WIDGET_HAS_FOCUS(this->entry_scale)) {
      gtk_entry_set_text(this->entry_scale, format_difftime((my_time_t)newval, 1));
   }

   this->num_changes++;
}

static void entry_scale__activate(GtkEntry *entry, gpointer user_data) {
//   THIS;

/* ##TODO##
   my text = this->entry_scale->get_text();
   my newval = 0;
   if(text =~ m!^\s*(\d+)\s*d(?:ays?)?(.*)!) {
      newval += $1 * 86400;
      text = $2;
   }
   if(text =~ m!^\s*(\d+)\s*h(?:ours?)?(.*)!) {
      newval += $1 * 3600;
      text = $2;
   }
   if(text =~ m!^\s*(\d+)\s*m(?:in(?:ute)?s?)?(.*)!) {
      newval += $1 * 60;
      text = $2;
   }
   if(text =~ m!^\s*(\d+)\s*s(?:ec(?:ond)?s?)?(.*)!) {
      newval += $1;
      text = $2;
   }
   if(text =~ m!^\s*(\d+)\s*:\s*(\d+)\s*:\s*(\d+)\s*!) {
      newval += $1 * 3600 + $2 * 60 + $3;
   }
   my min = this->min_interval};
   my max = this->max_value} - this->min_value};
   newval = min if(newval < min);
   newval = max if(newval > max);
   this->set_page_length(newval);

   // below is copypasted from above

   my s = newval % 60; newval = int(newval / 60);
   my m = newval % 60; newval = int(newval / 60);
   my h = newval % 24; newval = int(newval / 24);

   this->entry_scale->set_text(sprintf("%d days %02d:%02d:%02d", newval, h, m, s));
*/
}


data_scrollbar_controller_t *data_scrollbar_controller_new(GladeXML *gladexml, const char *widget_prefix) {
   data_scrollbar_controller_t *this = zmalloc_c(sizeof(data_scrollbar_controller_t));
   this->min_interval = 1;
   this->scroll_off_rel_pos = NAN;

   size_t wpl = strlen(widget_prefix);
   char tmp[wpl + 18];
   memcpy(tmp, widget_prefix, wpl);

   GtkRange *scroll_scale   = GTK_RANGE(glade_xml_get_widget(gladexml, (strcpy(tmp+wpl, "scrollbar_scale"), tmp)));
   GtkRange *scroll_off     = GTK_RANGE(glade_xml_get_widget(gladexml, (strcpy(tmp+wpl, "scrollbar_off"), tmp)));
   GtkRange *scroll_off_rel = GTK_RANGE(glade_xml_get_widget(gladexml, (strcpy(tmp+wpl, "scrollbar_off_rel"), tmp)));
   GtkRange *scale          = GTK_RANGE(glade_xml_get_widget(gladexml, (strcpy(tmp+wpl, "scale"), tmp)));

   GtkEntry *entry_scale    = GTK_ENTRY(glade_xml_get_widget(gladexml, (strcpy(tmp+wpl, "entry_scale"), tmp)));
   GtkLabel *label_off_min  = GTK_LABEL(glade_xml_get_widget(gladexml, (strcpy(tmp+wpl, "label_off_min"), tmp)));
   GtkLabel *label_off_max  = GTK_LABEL(glade_xml_get_widget(gladexml, (strcpy(tmp+wpl, "label_off_max"), tmp)));

   this->scroll_scale = scroll_scale;
   this->scroll_off = scroll_off;
   this->scale = scale;
   this->entry_scale = entry_scale;
   this->label_off_min = label_off_min;
   this->label_off_max = label_off_max;

   GtkAdjustment *hsa_scale = gtk_range_get_adjustment(scroll_scale);
   hsa_scale->step_increment = log(2.0) / 10; // double/halves the scale in 10 presses if "step" change
   hsa_scale->page_increment = log(2.0); // doubles/halves the scale if "page" change
   hsa_scale->page_size = 0;
   gtk_adjustment_changed(hsa_scale);

   GtkAdjustment *hsa_off = gtk_range_get_adjustment(scroll_off);
   hsa_off->step_increment = 10.0; // ##TODO## configure
   hsa_off->page_increment = 600.0;
   hsa_off->page_size = 0;
   gtk_adjustment_changed(hsa_off);

   if(scroll_off_rel) {

      // by using discontinuous mode, the value_changed is just
      // called once after a mouse scrollbar drag while change_value
      // is called during the drag
      gtk_range_set_update_policy(scroll_off_rel, GTK_UPDATE_DISCONTINUOUS);

      // this is called when user lets go of the scrollbar
      g_signal_connect(scroll_off_rel, "value-changed", G_CALLBACK(scroll_off_rel__value_changed), this);

      // this is called when the user scrolls
      g_signal_connect(scroll_off_rel, "change-value", G_CALLBACK(scroll_off_rel__change_value), this);
   }

   // off changed
   g_signal_connect(scroll_off, "value-changed", G_CALLBACK(scroll_off__value_changed), this);

   // scale changed
   g_signal_connect(scroll_scale, "value-changed", G_CALLBACK(scroll_scale__value_changed), this);

   if(entry_scale) {
      // if user edits the entry_scale field and hits enter, try to parse it!
      g_signal_connect(entry_scale, "activate", G_CALLBACK(entry_scale__activate), this);
   }

   return this;
}

void data_scrollbar_controller_set_page_off(data_scrollbar_controller_t *this, double off) {
   gtk_range_set_value(this->scroll_off, off);
}

double data_scrollbar_controller_get_page_off(data_scrollbar_controller_t *this) {
   return gtk_range_get_value(this->scroll_off);
}

void data_scrollbar_controller_set_page_length(data_scrollbar_controller_t *this, double length) {
   if(length > this->max_value - this->min_value) {
      length = this->max_value - this->min_value;
   } else if(length < this->min_interval) {
      length = this->min_interval;
   }
   gtk_range_set_value(this->scroll_scale, log(length));
}

void data_scrollbar_controller_set_page_length_log(data_scrollbar_controller_t *this, double loglength) {
   if(loglength > log(this->max_value - this->min_value)) {
      loglength = log(this->max_value - this->min_value);
   } else if(loglength < log(this->min_interval)) {
      loglength = log(this->min_interval);
   }
   gtk_range_set_value(this->scroll_scale, loglength);
}

double data_scrollbar_controller_get_page_length(data_scrollbar_controller_t *this) {
   return exp(gtk_range_get_value(this->scroll_scale));
}

double data_scrollbar_controller_get_page_length_log(data_scrollbar_controller_t *this) {
   return gtk_range_get_value(this->scroll_scale);
}

void data_scrollbar_controller_set_range(data_scrollbar_controller_t *this, double lo, double hi) {
   this->min_value = lo;
   this->max_value = hi;
   gtk_range_set_range(this->scroll_off, this->min_value, this->max_value);
   gtk_range_set_range(this->scroll_scale, log(this->min_interval), log(this->max_value - this->min_value));
}

void data_scrollbar_controller_get_range(data_scrollbar_controller_t *this, double *lo, double *hi) {
   *lo = this->min_value;
   *hi = this->max_value;
}

void data_scrollbar_controller_set_min_interval(data_scrollbar_controller_t *this, double min_interval) {
   this->min_interval = min_interval;
   gtk_range_set_range(this->scroll_scale, log(this->min_interval), log(this->max_value - this->min_value));
}

double data_scrollbar_controller_get_min_interval(data_scrollbar_controller_t *this) {
   return this->min_interval;
}

void data_scrollbar_controller_set_scale_pos(data_scrollbar_controller_t *this, double scale_pos) {
   if(!this->scale) return;
   if(scale_pos < 0) {
      printf("warning: scale_pos < 0 (%lf)\n", scale_pos);
      scale_pos = 0;
   } else if(scale_pos > 1) {
      printf("warning: scale_pos > 1 (%lf)\n", scale_pos);
      scale_pos = 1;
   }
   gtk_range_set_value(this->scale, scale_pos);
}

double data_scrollbar_controller_get_scale_pos(data_scrollbar_controller_t *this) {
   if(!this->scale) return 0;
   return gtk_range_get_value(this->scale);
}
