/*
 * Plot - plot time-based data on screen and to file with interactive controls
 * Copyright (C) 2006  Jonas Berlin <xkr47@outerspace.dyndns.org>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ****
 * plot2 - most functions related to representing & managing the
 * actual graphs along with gtk event handlers - except for the
 * plotting itself, see plot_draw
 */

#include <ctype.h>
#include <stdio.h>
#include <gtk/gtk.h>
#include <glade/glade.h>
#include <time.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "gnu_extensions.h"
#include "stringutil.xh"
#include "grid.xh"
#include "mem.xh"
#include "util.xh"
#include "plot2.xh"
#include "export.xh"
#include "plot_draw.xh"
#include "array.xh"

static void plot_need_redraw(my_plot_t *plot);

static void about_clicked(GtkMenuItem *about_menuitem, my_plot_callback_data_t *data) {
   glade_xml_new(data->gladefile, "aboutdialog1", NULL);
}

const char *labelNames[] = {
   "File date:",
   "Start time:",
   "End time:",
   "Covered time:",
   "Time format:",
   "Data columns:",
};

typedef struct {
      GtkLabel *labels[SZ(labelNames)];
} adf_data_t;

static void adf_update(GtkFileChooser *chooser, adf_data_t *data) {
   char *filename = gtk_file_chooser_get_preview_filename(chooser);
   my_dataset_t *dataset = scan_dataset(filename, 10);
   g_free(filename);
   gtk_file_chooser_set_preview_widget_active(chooser, 1);
   if(!dataset) {
      for(int i=0; i<SZ(labelNames); ++i) {
	 gtk_label_set_text(data->labels[i], "?");
      }
      return;
   }

   char buf[40];
   int idx = 0;

   gtk_label_set_text(data->labels[idx++], format_time(dataset->last_modified, 0));
   gtk_label_set_text(data->labels[idx++], format_time(dataset->ts, 0));
   gtk_label_set_text(data->labels[idx++], format_time(dataset->te, 0));
   gtk_label_set_text(data->labels[idx++], format_difftime(dataset->te - dataset->ts, 0));
   gtk_label_set_text(data->labels[idx++], dataset->timetype == TIMETYPE_EPOCH ? "epoch" : "relative");
   snprintf(buf, sizeof(buf), "%d", dataset->item_count);
   gtk_label_set_text(data->labels[idx++], buf);

   unref_dataset(dataset);
}


static void add_data_file_clicked(GtkMenuItem *menuitem, my_plot_callback_data_t *data) {

   GtkWidget* dialog = gtk_file_chooser_dialog_new("Add datafile",
						   data->parent,
						   GTK_FILE_CHOOSER_ACTION_OPEN,
						   "gtk-cancel", GTK_RESPONSE_CANCEL,
						   "gtk-open", GTK_RESPONSE_ACCEPT,
						   NULL);

   GtkFileChooser *chooser = GTK_FILE_CHOOSER(dialog);

   gtk_file_chooser_set_select_multiple(chooser, 1);

   GtkWidget *table = gtk_table_new(5, 2, 0);
   gtk_table_set_col_spacings(GTK_TABLE(table), 2);

   adf_data_t adtd;

   for(unsigned i=0; i<SZ(labelNames); ++i) {
      GtkWidget *labelNameW = gtk_label_new(labelNames[i]);
      GtkWidget *labelW = gtk_label_new("?");

      gtk_misc_set_alignment(GTK_MISC(labelNameW), 1.0, 0.5);
      gtk_misc_set_alignment(GTK_MISC(labelW), 0.0, 0.5);

      adtd.labels[i] = GTK_LABEL(labelW);

      gtk_table_attach(GTK_TABLE(table), labelNameW, 0, 1, i, i+1, GTK_EXPAND|GTK_FILL, 0, 0, 0);
      gtk_table_attach(GTK_TABLE(table), labelW,     1, 2, i, i+1, GTK_EXPAND|GTK_FILL, 0, 0, 0);
   }
   gtk_widget_show_all(table);

   gtk_file_chooser_set_preview_widget(chooser, table);
   g_signal_connect(chooser, "update-preview", G_CALLBACK(adf_update), &adtd);

   GtkFileFilter *filter;

   filter = gtk_file_filter_new();
   gtk_file_filter_set_name(filter, "Text files");
   gtk_file_filter_add_mime_type(filter, "text/plain");
   gtk_file_chooser_add_filter(chooser, filter);

   filter = gtk_file_filter_new();
   gtk_file_filter_set_name(filter, "All files");
   gtk_file_filter_add_pattern(filter, "*");
   gtk_file_chooser_add_filter(chooser, filter);

   if(gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT) {
      GSList* files;
      for(files = gtk_file_chooser_get_filenames(chooser); files; files = files->next) {
	 add_datafile(data->plot, files->data);
      }
      g_slist_free (files);
   }

   gtk_widget_destroy (dialog);
}

void free_plot_line_info(my_plot_line_info_t *p) {
   if(p->line_gc) g_object_unref(p->line_gc);
   if(p->horiz_marker) g_object_unref(p->horiz_marker);
   if(p->color) free(p->color);
}

void free_plot_line_info_set(my_plot_line_info_set_t *p) {
   for(int i=0; i<p->line_count; ++i) {
      free_plot_line_info(&p->line[i]);
   }
   my_plot_line_info_arr_free(p->line);

   unref_dataset(p->dataset);
}

void remove_data_file(my_plot_t *plot, int idx) {
   free_plot_line_info_set(&plot->line_set[idx]);
   --plot->line_set_count;
   printf("%p %p %ld\n", &plot->line_set[idx], &plot->line_set[idx+1], (long)(plot->line_set_count - idx) * sizeof(my_plot_line_info_set_t));
   memmove(&plot->line_set[idx], &plot->line_set[idx+1], (plot->line_set_count - idx) * sizeof(my_plot_line_info_set_t));

   plot_need_redraw(plot);
   update_minmax(plot);
}

void remove_all_data_files(my_plot_t *plot) {
   for(int i=0; i<plot->line_set_count; ++i) {
      free_plot_line_info_set(&plot->line_set[i]);
   }
   my_plot_line_info_set_arr_free(plot->line_set);

   plot_need_redraw(plot);
   update_minmax(plot);
}

static void remove_data_file_clicked(GtkMenuItem *menuitem, my_plot_callback_data_t *data) {
   GladeXML *gladexml = glade_xml_new(data->gladefile, "dialog_remove_datafile", NULL);
   GtkWidget *dialog = glade_xml_get_widget(gladexml, "dialog_remove_datafile");

   GtkWidget *combo = glade_xml_get_widget(gladexml, "combobox_datafiles");

   gtk_combo_box_remove_text(GTK_COMBO_BOX(combo), 0);

   for(int i=0; i<data->plot->line_set_count; ++i) {
      gtk_combo_box_append_text(GTK_COMBO_BOX(combo), data->plot->line_set[i].dataset->path);
   }

   if(gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_OK) {
      int idx = gtk_combo_box_get_active(GTK_COMBO_BOX(combo));
      if(idx >= 0) {
// 2.6 or 2.8 ?
#if GTK_CHECK_VERSION(2,6,0)
	 printf("removing %d. %s\n", idx, gtk_combo_box_get_active_text(GTK_COMBO_BOX(combo)));
#else
	 printf("removing %d.\n", idx);
#endif
	 remove_data_file(data->plot, idx);
      }
   }

   gtk_widget_destroy (dialog);
}

static void remove_all_data_files_clicked(GtkMenuItem *menuitem, my_plot_callback_data_t *data) {
   remove_all_data_files(data->plot);
}

void plot_init(my_plot_t *plot) {
   bzero(plot, sizeof(my_plot_t));
   plot->line_set_dflt.line_dflt.hline = -1;
   plot->line_set_dflt.line_dflt.dataset_idx = -1;
}

void plot_release(my_plot_t *plot) {
   if(plot->minor_gc) {
      g_object_unref(plot->minor_gc);
      g_object_unref(plot->major_gc);
      g_object_unref(plot->sep_gc);
      g_object_unref(plot->label_gc);
      g_object_unref(plot->cursor_marker_gc);
   }
}

// also see DA_ALLOCATION in plot_draw.xc
#define DA_ALLOCATION int w = plot->da->allocation.width; int h = plot->da->allocation.height; h -= 24;


static gint8 cursor_marker_dashes[] = { 2, 4 };

static void plot_need_redraw(my_plot_t *plot) {
   //printf("need_redraw\n");
   if(plot->conc == 0) {
      gtk_widget_queue_draw(plot->da);
   } else {
      plot->normal_expose++;
   }
}

static void add_to_zoom_history(my_plot_t *plot) {
   int idx = plot->zoom_pos + 1;
   if(idx > plot->zoom_history_count) idx = plot->zoom_history_count;

   my_zoom_view_arr_ensure(plot->zoom_history, idx + 1);
   my_zoom_view_t *z = &plot->zoom_history[idx];

   z->xoff = data_scrollbar_controller_get_page_off(plot->dsc_x);
   z->yoff = data_scrollbar_controller_get_page_off(plot->dsc_y);
   z->xlength_log = data_scrollbar_controller_get_page_length_log(plot->dsc_x);
   z->ylength_log = data_scrollbar_controller_get_page_length_log(plot->dsc_y);
   z->xscalepos = data_scrollbar_controller_get_scale_pos(plot->dsc_x);
   z->yscalepos = data_scrollbar_controller_get_scale_pos(plot->dsc_y);

   if(idx > 0) {
      gtk_widget_set_sensitive(plot->button_zoom_prev, 1);
   }
   gtk_widget_set_sensitive(plot->button_zoom_next, 0);

   plot->zoom_pos = idx;
}

static void switch_to_zoom_history(my_plot_t *plot, int idx) {
   if(idx < 0) return;
   if(idx >= plot->zoom_history_count) return;
   if(idx == plot->zoom_pos) return;

   plot->zoom_pos = idx;

   gtk_widget_set_sensitive(plot->button_zoom_prev, idx > 0);
   gtk_widget_set_sensitive(plot->button_zoom_next, idx < plot->zoom_history_count-1);

   my_zoom_view_t *z = &plot->zoom_history[idx];

   data_scrollbar_controller_set_page_length_log(plot->dsc_x, z->xlength_log);
   data_scrollbar_controller_set_page_length_log(plot->dsc_y, z->ylength_log);
   data_scrollbar_controller_set_page_off(plot->dsc_x, z->xoff);
   data_scrollbar_controller_set_page_off(plot->dsc_y, z->yoff);
   data_scrollbar_controller_set_scale_pos(plot->dsc_x, z->xscalepos);
   data_scrollbar_controller_set_scale_pos(plot->dsc_y, z->yscalepos);
}

static void show_zoom_history(my_plot_t *plot, int idx) {
   DA_ALLOCATION;

   if(idx < 0) return;
   if(idx >= plot->zoom_history_count) return;

   my_zoom_view_t *z = &plot->zoom_history[idx];

   double xoff = data_scrollbar_controller_get_page_off(plot->dsc_x);
   double yoff = data_scrollbar_controller_get_page_off(plot->dsc_y);
   double xlength = data_scrollbar_controller_get_page_length(plot->dsc_x);
   double ylength = data_scrollbar_controller_get_page_length(plot->dsc_y);

   double z__xoff2 = z->xoff + exp(z->xlength_log);
   double z__yoff2 = z->yoff + exp(z->ylength_log);

   int x = (int)(w * (z->xoff - xoff) / xlength);
   int y = (int)(h * (z->yoff - yoff) / ylength);
   int x2 = (int)(w * (z__xoff2 - xoff) / xlength);
   int y2 = (int)(h * (z__yoff2 - yoff) / ylength);

   y = h-1 - y;
   y2 = h-1 - y2;

   plot->bx = x;
   plot->by = y;
   plot->box = x2;
   plot->boy = y2;

   gdk_gc_set_dashes(plot->cursor_marker_gc, -plot->hline, cursor_marker_dashes, SZ(cursor_marker_dashes));
   x_gdk_draw_rectangle(plot->da->window, plot->cursor_marker_gc, 0, x, y, x2, y2);
}

static void hide_zoom_history(my_plot_t *plot) {
   if(plot->box == COORD_UNDEF) return;

   gdk_gc_set_dashes(plot->cursor_marker_gc, -plot->hline, cursor_marker_dashes, SZ(cursor_marker_dashes));
   x_gdk_draw_rectangle(plot->da->window, plot->cursor_marker_gc, 0, plot->bx, plot->by, plot->box, plot->boy);

   plot->box = COORD_UNDEF;
}

static void button_zoom_prev_clicked(GtkButton *button, my_plot_t *plot) {
   switch_to_zoom_history(plot, plot->zoom_pos - 1);
}

static void button_zoom_next_clicked(GtkButton *button, my_plot_t *plot) {
   switch_to_zoom_history(plot, plot->zoom_pos + 1);
}

static void button_zoom_hover(GtkButton *button, my_plot_t *plot) {
   show_zoom_history(plot, plot->zoom_pos + (GTK_WIDGET(button) == plot->button_zoom_prev ? -1 : 1));
}

static void button_zoom_unhover(GtkButton *button, my_plot_t *plot) {
   hide_zoom_history(plot);
}

static void xor_markers(my_plot_t *plot) {
   DA_ALLOCATION;

   if(plot->hline >= 0) {
      gdk_gc_set_dashes(plot->cursor_marker_gc, -plot->vline, cursor_marker_dashes, SZ(cursor_marker_dashes));
      gdk_draw_line(plot->da->window, plot->cursor_marker_gc, 0, plot->hline, w, plot->hline);
      gdk_gc_set_dashes(plot->cursor_marker_gc, -plot->hline, cursor_marker_dashes, SZ(cursor_marker_dashes));
      gdk_draw_line(plot->da->window, plot->cursor_marker_gc, plot->vline, 0, plot->vline, h);
   }

   for(int i=0; i<plot->line_set_count; ++i) {
      for(int j=0; j<plot->line_set[i].line_count; ++j) {
	 if(plot->line_set[i].line[j].hline >= 0) {
	    gdk_draw_line(plot->da->window,
			  plot->line_set[i].line[j].horiz_marker,
			  0, plot->line_set[i].line[j].hline,
			  w, plot->line_set[i].line[j].hline);
	 }
      }
   }
}

void hide_markers(my_plot_t *plot) {
   xor_markers(plot);

   plot->hline = -1;
   plot->vline = -1;
   for(int i=0; i<plot->line_set_count; ++i) {
      for(int j=0; j<plot->line_set[i].line_count; ++j) {
	 plot->line_set[i].line[j].hline = -1;
      }
   }
}

// TODO this is sometimes called befor GCs have been initialized because they are initialized in expose
static void show_markers(my_plot_t *plot, int x, int y) {
   DA_ALLOCATION;

   gdk_gc_set_dashes(plot->cursor_marker_gc, -x, cursor_marker_dashes, SZ(cursor_marker_dashes));
   gdk_draw_line(plot->da->window, plot->cursor_marker_gc, 0, y, w, y);
   gdk_gc_set_dashes(plot->cursor_marker_gc, -y, cursor_marker_dashes, SZ(cursor_marker_dashes));
   gdk_draw_line(plot->da->window, plot->cursor_marker_gc, x, 0, x, h);
   if(plot->hline >= 0) {
      gdk_gc_set_dashes(plot->cursor_marker_gc, -plot->vline, cursor_marker_dashes, SZ(cursor_marker_dashes));
      gdk_draw_line(plot->da->window, plot->cursor_marker_gc, 0, plot->hline, w, plot->hline);
      gdk_gc_set_dashes(plot->cursor_marker_gc, -plot->hline, cursor_marker_dashes, SZ(cursor_marker_dashes));
      gdk_draw_line(plot->da->window, plot->cursor_marker_gc, plot->vline, 0, plot->vline, h);
   }
   plot->vline = x;
   plot->hline = y;

   double xlength = data_scrollbar_controller_get_page_length(plot->dsc_x);
   double xoff = data_scrollbar_controller_get_page_off(plot->dsc_x);

   double ylength = data_scrollbar_controller_get_page_length(plot->dsc_y);
   double yoff = data_scrollbar_controller_get_page_off(plot->dsc_y);

   my_time_t xv = (my_time_t)(x * xlength / w + xoff);
   for(int i=0; i<plot->line_set_count; ++i) {
      int p = my_time_arr_find(plot->line_set[i].dataset->time, xv, 0);
      for(int j=0; j<plot->line_set[i].line_count; ++j) {
	 if(!plot->line_set[i].line[j].enabled) continue;

	 int dataset_idx = plot->line_set[i].line[j].dataset_idx;

	 int dy = !__isnan(plot->line_set[i].dataset->item[dataset_idx].data[p]) ?
	    (int)(h - (plot->line_set[i].dataset->item[dataset_idx].data[p] - yoff) / ylength * h) :
	    -1;

	 if(dy >= 0) gdk_draw_line(plot->da->window, plot->line_set[i].line[j].horiz_marker, 0, dy, w, dy);

	 if(plot->line_set[i].line[j].hline >= 0) {
	    gdk_draw_line(plot->da->window,
			  plot->line_set[i].line[j].horiz_marker,
			  0, plot->line_set[i].line[j].hline,
			  w, plot->line_set[i].line[j].hline);
	 }

	 plot->line_set[i].line[j].hline = dy;
      }
   }
}

static void plot_move_to(my_plot_t *plot, int x, int y) {
   DA_ALLOCATION;

/*
  data_scrollbar_controller_set_scale_pos(plot->dsc_x, ((double)x) / w);
  data_scrollbar_controller_set_scale_pos(plot->dsc_y, 1 - ((double)y) / h);
*/

   int xd = x - plot->bx;
   int yd = plot->by - y;

   xd *= plot->bspeed;
   yd *= plot->bspeed;

   //printf("move %d,%d\n", xd, yd);

   double xmin, xmax;
   double ymin, ymax;

   data_scrollbar_controller_get_range(plot->dsc_x, &xmin, &xmax);
   data_scrollbar_controller_get_range(plot->dsc_y, &ymin, &ymax);

   double xlength = data_scrollbar_controller_get_page_length(plot->dsc_x);
   double ylength = data_scrollbar_controller_get_page_length(plot->dsc_y);

   double nxoff = plot->bxoff - xd * xlength / w;
   double nyoff = plot->byoff - yd * ylength / h;

   //printf("nt %lf %lf\n", nxoff, nyoff);

   if(nxoff < xmin) nxoff = xmin; else if(nxoff + xlength > xmax) nxoff = xmax - xlength;
   if(nyoff < ymin) nyoff = ymin; else if(nyoff + ylength > ymax) nyoff = ymax - ylength;

   //printf("n %lf %lf\n", nxoff, nyoff);

   data_scrollbar_controller_set_page_off(plot->dsc_x, nxoff);
   data_scrollbar_controller_set_page_off(plot->dsc_y, nyoff);
}

static int inside_range(int x, int min, int max) {
   return x >= min && x <= max;
}

static int check_if_hit_legend_box(my_plot_t *plot, int x, int y) {
   return
      inside_range(x, plot->legend_x, plot->legend_x2) &&
      inside_range(y, plot->legend_y, plot->legend_y2);
}

static my_plot_line_info_t *check_if_hit_legend_item(my_plot_t *plot, int x, int y) {
   for(int lsi = 0; lsi < plot->line_set_count; ++lsi) {
      my_plot_line_info_t *line = plot->line_set[lsi].line;
      int line_count = plot->line_set[lsi].line_count;

      for(int i=0; i<line_count; ++i) {
	 if(inside_range(x, line[i].legend_x, line[i].legend_x2) &&
	    inside_range(y, line[i].legend_y, line[i].legend_y2)) {
	    return &line[i];
	 }
      }
   }

   return NULL;
}

static gboolean plot_button_press_event(GtkWidget *widget, GdkEventButton *event, my_plot_t *plot) {
   int x = event->x;
   int y = event->y;

   switch((int)event->type) {
/*
  case GDK_2BUTTON_PRESS: {
  DA_ALLOCATION;

  printf("dpress %d %d\n", x, y);
  plot_move_to(plot, w/2, h/2);
  break;
  }
*/

      case GDK_BUTTON_PRESS:
	 // printf("press %d %d\n", x, y);
	 if(plot->bmode == 0) {
	    DA_ALLOCATION;

	    plot->bx = x;
	    plot->by = y;

	    hide_markers(plot);

	    if(check_if_hit_legend_box(plot, x, y)) {
	       my_plot_line_info_t *line = check_if_hit_legend_item(plot, x, y);

	       if(line) {
		  switch(event->button) {
		     case 1:
			plot->bstate = line->enabled = !line->enabled;
			plot->bmode = 4;

			plot_need_redraw(plot);
			update_minmax(plot);
			break;

			/*
		     default:
			diff mode, TODO scales don't work and we need a time unit also
			plot->bstate = line->mode = !line->mode;
			plot->bmode = 5;

			plot_need_redraw(plot);
			update_minmax(plot);
			break;
			*/
		  }
	       }
	       return 0;
	    }

	    // update scale pos according to press location
	    data_scrollbar_controller_set_scale_pos(plot->dsc_x, ((double)x) / w);
	    data_scrollbar_controller_set_scale_pos(plot->dsc_y, 1 - ((double)y) / h);

	    plot->bspeed = event->state & GDK_SHIFT_MASK ? 4 : 1;

	    if(event->button > 1) {
	       plot->bmode = 2;
	       plot->bxloglength = data_scrollbar_controller_get_page_length_log(plot->dsc_x);
	       plot->byloglength = data_scrollbar_controller_get_page_length_log(plot->dsc_y);

	       // when close to border, zoom to border
	       if(x < 30) { // ##TODO## conf 30
		  data_scrollbar_controller_set_scale_pos(plot->dsc_x, 0.0);
	       } else if(x >= w - 30) {
		  data_scrollbar_controller_set_scale_pos(plot->dsc_x, 1.0);
	       }

	       if(y < 30) {
		  data_scrollbar_controller_set_scale_pos(plot->dsc_y, 1.0);
	       } else if(y >= h - 30) {
		  data_scrollbar_controller_set_scale_pos(plot->dsc_y, 0.0);
	       }

	    } else if(event->state & GDK_CONTROL_MASK) {
	       plot->bmode = 3;
	       plot->box = COORD_UNDEF;

	    } else {
	       plot->bmode = 1;
	       plot->bxoff = data_scrollbar_controller_get_page_off(plot->dsc_x);
	       plot->byoff = data_scrollbar_controller_get_page_off(plot->dsc_y);
	    }
	 }
	 break;
   }

   return 0;
}

#define LIMIT(v,min,max) do{ if(v<(min)) v=min; else if(v>(max)) v=max; } while(0)

static void plot_motion_generic(my_plot_t *plot, int x, int y, int release) {
   switch(plot->bmode) {
      case 0:
	 show_markers(plot, x, y);
	 break;

      case 1:
	 plot_move_to(plot, x, y);
	 break;

      case 2: {
	 DA_ALLOCATION;

	 // ##TODO## conf 5.0 and 2.0
	 data_scrollbar_controller_set_page_length_log(plot->dsc_x, plot->bxloglength - (x - plot->bx) * plot->bspeed * 5.0 / w);
	 data_scrollbar_controller_set_page_length_log(plot->dsc_y, plot->byloglength - (y - plot->by) * plot->bspeed * 2.0 / h);
	 break;
      }

      case 3: {
	 DA_ALLOCATION;

	 w--;
	 h--;

	 LIMIT(plot->bx, 0, w);
	 LIMIT(plot->by, 0, h);
	 LIMIT(x, 0, w);
	 LIMIT(y, 0, h);

	 gdk_gc_set_dashes(plot->cursor_marker_gc, -plot->hline, cursor_marker_dashes, SZ(cursor_marker_dashes));
	 if(!release) {
	    x_gdk_draw_rectangle(plot->da->window, plot->cursor_marker_gc, 0, plot->bx, plot->by, x, y);
	 }
	 if(plot->box != COORD_UNDEF) {
	    x_gdk_draw_rectangle(plot->da->window, plot->cursor_marker_gc, 0, plot->bx, plot->by, plot->box, plot->boy);
	 }
	 if(!release) {
	    plot->box = x;
	    plot->boy = y;
	 } else {
	    int x1 = plot->bx;
	    int y1 = plot->by;

	    y = h - y;
	    y1 = h - y1;

	    if(x < x1) { int t = x; x = x1; x1 = t; }
	    if(y < y1) { int t = y; y = y1; y1 = t; }

	    double ox = data_scrollbar_controller_get_page_off(plot->dsc_x);
	    double oy = data_scrollbar_controller_get_page_off(plot->dsc_y);
	    double lx = data_scrollbar_controller_get_page_length(plot->dsc_x);
	    double ly = data_scrollbar_controller_get_page_length(plot->dsc_y);

	    //printf("%d,%d %d,%d\n", x1, y1, x-x1, y-y1);

	    //printf("%lf,%lf %lf,%lf\n", ox, oy, lx, ly);

	    ox += lx * x1 / w;
	    oy += ly * y1 / h;

	    lx *= (x - x1) / (double)w;
	    ly *= (y - y1) / (double)h;

	    //printf("%lf,%lf %lf,%lf\n\n", ox, oy, lx, ly);

	    // after area zoom, always put center as zoom point
	    data_scrollbar_controller_set_scale_pos(plot->dsc_x, 0.5);
	    data_scrollbar_controller_set_scale_pos(plot->dsc_y, 0.5);

	    // update page lengths
	    data_scrollbar_controller_set_page_length(plot->dsc_x, lx);
	    data_scrollbar_controller_set_page_length(plot->dsc_y, ly);

	    // now, in case either page length subceeded the minimum page length, we adjust offset so that the selected zoom area becomes centered in view
	    double nlx = data_scrollbar_controller_get_page_length(plot->dsc_x);
	    double nly = data_scrollbar_controller_get_page_length(plot->dsc_y);

	    ox -= (nlx - lx) / 2;
	    oy -= (nly - ly) / 2;

	    // update offsets
	    data_scrollbar_controller_set_page_off(plot->dsc_x, ox);
	    data_scrollbar_controller_set_page_off(plot->dsc_y, oy);

	    //printf("%lf,%lf %lf,%lf\n\n", data_scrollbar_controller_get_page_off(plot->dsc_x), data_scrollbar_controller_get_page_off(plot->dsc_y), nlx, nly);

	    plot->box = COORD_UNDEF;
	 }
	 break;
      }

      case 4: {
	 if(!check_if_hit_legend_box(plot, x, y)) break;

	 my_plot_line_info_t *line = check_if_hit_legend_item(plot, x, y);

	 if(line) {
	    int changed = line->enabled != plot->bstate;
	    if(changed) {
	       line->enabled = plot->bstate;

	       plot_need_redraw(plot);
	       update_minmax(plot);
	    }
	 }
	 break;
      }

      case 5: {
	 if(!check_if_hit_legend_box(plot, x, y)) break;

	 my_plot_line_info_t *line = check_if_hit_legend_item(plot, x, y);

	 if(line) {
	    int changed = line->mode != plot->bstate;
	    if(changed) {
	       line->mode = plot->bstate;

	       plot_need_redraw(plot);
	       update_minmax(plot);
	    }
	 }
	 break;
      }

      default:
	 fprintf(stderr, "ERROR: bmode = %d\n", plot->bmode);
	 break;
   }
}

static gboolean plot_button_release_event(GtkWidget *widget, GdkEventButton *event, my_plot_t *plot) {
   int x = event->x;
   int y = event->y;

   plot_motion_generic(plot, x, y, 1);

   plot->bmode = 0;

   add_to_zoom_history(plot);

   return 0;
}

static gboolean plot_leave_notify_event(GtkWidget *widget, GdkEventCrossing *event, my_plot_t *plot) {
   hide_markers(plot);

   return 0;
}

static gboolean plot_motion_notify_event(GtkWidget *wid, GdkEventMotion *event, my_plot_t *plot) {
   int x = event->x;
   int y = event->y;

   plot_motion_generic(plot, x, y, 0);

   return 0;
}

/*
  my $co_minor = $#colors - 3;
  my $co_major = $#colors - 2;
  my $co_sep = $#colors - 1;
  my $co_label = $#colors - 0;
*/

static void plot_scrollbar_value_changed(GtkRange *range, my_plot_t *plot) {
   plot_need_redraw(plot);
}

static gboolean plot_configure_event(GtkWidget *widget, GdkEventConfigure *do_not_use, my_plot_t *plot) {
   DA_ALLOCATION;

   //printf("configure\n");

   data_scrollbar_controller_set_min_interval(plot->dsc_x, (double)w / 2);
   data_scrollbar_controller_set_min_interval(plot->dsc_y, (double)1);

   return 0;
}

static gboolean plot_map_event(GtkWidget *widget, GdkEvent *do_not_use, my_plot_t *plot) {
   //printf("map\n");

   // set the background part 1 - create new style for widget
   GtkStyle *style = gtk_widget_get_style(plot->da);
   style = gtk_style_copy(style);
   gtk_widget_set_style(plot->da, style);

   // set the background part 2
   GdkColor bg;
   get_color(&bg, "#000000");
   gdk_window_set_background(plot->da->window, &bg);

   plot_configure_event(widget, NULL, plot);

   return 0;
}

static gboolean plot_expose_event(GtkWidget *widget, GdkEventExpose *event, my_plot_t *plot) {
   //printf("expose\n");

   if(plot->conc > 0) {
      // still not optimal
      plot->normal_expose++;
      //printf(" -> normal_expose++\n");
      return 0;
   }
   int mytrans = ++plot->dsc_x->num_changes + plot->dsc_y->num_changes;
   plot->conc++;
   //    print "->$mytrans plot->conc\n";

   plot_draw_to(plot, plot->da->window, 1);

   if(mytrans != plot->dsc_x->num_changes + plot->dsc_y->num_changes || plot->normal_expose) {
      //printf("%d pending exposes, queue draw\n", plot->normal_expose);
      gtk_widget_queue_draw(plot->da);
      plot->normal_expose = 0;
   }
   plot->conc--;

   //printf("/expose\n");

   return 0;
}

void update_minmax(my_plot_t *plot) {
   my_time_t ts = 0, te = 1;
   double vmin = INFINITY, vmax = -INFINITY;

   int count = 0;

   for(int lsi=0; lsi<plot->line_set_count; ++lsi) {
      my_dataset_t *dataset = plot->line_set[lsi].dataset;
      my_plot_line_info_t *line = plot->line_set[lsi].line;
      int line_count = plot->line_set[lsi].line_count;
      for(int i=0; i<line_count; ++i) {
	 if(!line[i].enabled) continue;
	 int d = line[i].dataset_idx;
	 switch(line[i].mode) {
	    case 0:
	       if(count == 0 || vmin > dataset->item[d].min) vmin = dataset->item[d].min;
	       if(count == 0 || vmax < dataset->item[d].max) vmax = dataset->item[d].max;
	       break;
	    case 1:
	       if(count == 0 || vmin > dataset->item[d].min_diff) vmin = dataset->item[d].min_diff;
	       if(count == 0 || vmax < dataset->item[d].max_diff) vmax = dataset->item[d].max_diff;
	       break;
	 }
	 count++;
      }

      if(lsi == 0 || dataset->ts < ts) {
	 ts = dataset->ts;
      }

      if(lsi == 0 || dataset->te > te) {
	 te = dataset->te;
      }
   }

   // don't do anything if no lines visible
   if(count == 0) return;

   data_scrollbar_controller_set_range(plot->dsc_x, (double)ts, (double)te);
   data_scrollbar_controller_set_page_length(plot->dsc_x, (double)(te - ts)); // initially zoom out

   vmin -= (vmax-vmin)/20;
   vmax += (vmax-vmin)/20;

   data_scrollbar_controller_set_range(plot->dsc_y, vmin, vmax);
   data_scrollbar_controller_set_page_length(plot->dsc_y, vmax-vmin); // initially zoom out

   add_to_zoom_history(plot);
}

// ##TODO## make local
// ##TODO## the colors suck
const char *default_line_colors[] = { "#ff4444", "#ffc040", "#40ff40", "#40ffff", "#4080ff", "red", "green", "yellow", "blue", "white", "Wheat", "LemonChiffon", "BlueViolet", "dark orchid", "bisque2", "plum", "DarkOrange2" };
int linecoloff = 0;

void add_datafile(my_plot_t *plot, char *filename) {
   printf("Loading %s ..\n", filename);

   my_dataset_t *dataset = read_dataset(filename);

   if(!dataset) {
      fprintf(stderr, "\033[31;1mWARNING: File '%s' could not be loaded\033[m\n", filename);
      return;
   }

   dataset->maxdiff = 100 * TIME_MULTIPLIER;

   int setidx = plot->line_set_count;

   my_plot_line_info_set_arr_ensure(plot->line_set, setidx + 1);

   plot->line_set[setidx].dataset = dataset;

   for(int i=0; i<dataset->item_count; ++i) {
      int lineidx = plot->line_set[setidx].line_count;
      my_plot_line_info_arr_ensure(plot->line_set[setidx].line, lineidx+1);
      int linecolidx = linecoloff++;
      plot->line_set[setidx].line[lineidx].color = strdup(default_line_colors[linecolidx % SZ(default_line_colors)]);
      plot->line_set[setidx].line[lineidx].dataset_idx = i;
      plot->line_set[setidx].line[lineidx].enabled = 1;
      printf("%d %lf - %lf\n", i, dataset->item[i].min, dataset->item[i].max);
   }

   printf("Datafile %d: %d data items, %d used\n\n", setidx + 1, dataset->time_count, plot->line_set[setidx].line_count);

   plot_need_redraw(plot);
   update_minmax(plot);
}

gboolean destroy_event(GtkWidget *widget, GdkEvent *event, gpointer user_data);

void plot_connect_signals(my_plot_t *plot, GtkWindow *win, const char *gladefile, GladeXML *gladexml) {

   my_plot_callback_data_t *cbd = malloc(sizeof(my_plot_callback_data_t));
   cbd->parent = win;
   cbd->plot = plot;
   cbd->gladefile = gladefile;

   gtk_widget_add_events(plot->da, GDK_LEAVE_NOTIFY_MASK);
   g_signal_connect(plot->da, "leave-notify-event", G_CALLBACK(plot_leave_notify_event), plot);

   gtk_widget_add_events(plot->da, GDK_POINTER_MOTION_MASK);
   g_signal_connect(plot->da, "motion-notify-event", G_CALLBACK(plot_motion_notify_event), plot);

   //gtk_widget_add_events(plot->da, GDK_BUTTON_MOTION_MASK);
   //g_signal_connect(plot->da, "-event", G_CALLBACK(plot_motion_notify_event), plot);

   gtk_widget_add_events(plot->da, GDK_BUTTON_PRESS_MASK);
   g_signal_connect(plot->da, "button-press-event", G_CALLBACK(plot_button_press_event), plot);

   gtk_widget_add_events(plot->da, GDK_BUTTON_RELEASE_MASK);
   g_signal_connect(plot->da, "button-release-event", G_CALLBACK(plot_button_release_event), plot);


   g_signal_connect(plot->da, "map-event", G_CALLBACK(plot_map_event), plot);
   g_signal_connect(plot->da, "expose-event", G_CALLBACK(plot_expose_event), plot);
   g_signal_connect(plot->da, "configure-event", G_CALLBACK(plot_configure_event), plot);

   // buttons

   g_signal_connect(plot->button_zoom_prev, "clicked", G_CALLBACK(button_zoom_prev_clicked), plot);
   g_signal_connect(plot->button_zoom_next, "clicked", G_CALLBACK(button_zoom_next_clicked), plot);
   g_signal_connect(plot->button_zoom_prev, "enter", G_CALLBACK(button_zoom_hover), plot);
   g_signal_connect(plot->button_zoom_next, "enter", G_CALLBACK(button_zoom_hover), plot);
   g_signal_connect(plot->button_zoom_prev, "leave", G_CALLBACK(button_zoom_unhover), plot);
   g_signal_connect(plot->button_zoom_next, "leave", G_CALLBACK(button_zoom_unhover), plot);

   //g_signal_connect(glade_xml_get_widget(gladexml, "button_menu"), "clicked", G_CALLBACK(menu_clicked), glade_xml_get_widget(gladexml, "menubar_main"));

   // menu

   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_file_add_data_file"), "activate", G_CALLBACK(add_data_file_clicked), cbd);
   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_file_remove_data_file"), "activate", G_CALLBACK(remove_data_file_clicked), cbd);
   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_file_remove_all_data_files"), "activate", G_CALLBACK(remove_all_data_files_clicked), cbd);
   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_file_export_as"), "activate", G_CALLBACK(export_as_clicked), cbd);
   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_quit"), "activate", G_CALLBACK(destroy_event), (gpointer)2);

   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_view_previous"), "activate", G_CALLBACK(button_zoom_prev_clicked), plot);
   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_view_next"), "activate", G_CALLBACK(button_zoom_next_clicked), plot);

   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_about"), "activate", G_CALLBACK(about_clicked), cbd);

   // scrollbars

   g_signal_connect(glade_xml_get_widget(gladexml, "xscroll_scrollbar_off"), "value-changed", G_CALLBACK(plot_scrollbar_value_changed), plot);
   g_signal_connect(glade_xml_get_widget(gladexml, "xscroll_scrollbar_scale"), "value-changed", G_CALLBACK(plot_scrollbar_value_changed), plot);
   g_signal_connect(glade_xml_get_widget(gladexml, "yscroll_scrollbar_off"), "value-changed", G_CALLBACK(plot_scrollbar_value_changed), plot);
   g_signal_connect(glade_xml_get_widget(gladexml, "yscroll_scrollbar_scale"), "value-changed", G_CALLBACK(plot_scrollbar_value_changed), plot);
}
